import numpy as np
import matplotlib.pyplot as plt

# Objective function
def f(a0,a1,x,y):
    N=len(x)
    return 1.0/N*np.sum((a0+a1*x-y)**2)

def dfda0(a0,a1,x,y):
    N=len(x)
    return 2.0/N*np.sum( a0+a1*x-y)

def dfda1(a0,a1,x,y):
    N=len(x)
    return 2.0/N*np.sum((a0+a1*x-y)*x)

# Given data
x = np.array([1,2,3,4,5])
y = np.array([1,3,5,3,7])

N = len(x)

# Main
# Initialize
a0_0,a1_0 = 0.0,0.0
eps       = 1e-8

# Gradient descent method with Backtracking line search
a0_k,a1_k = a0_0,a1_0
iter      = 0
alpha,beta= 0.3,0.9

while True:
    # 1. Delta ak=-grad_f(ak)
    delta_a0_k=-dfda0(a0_k,a1_k,x,y)
    delta_a1_k=-dfda1(a0_k,a1_k,x,y)
    
    # 2. Compute tk with Backtracking line search
    tk=1
    while True:
        LHS = f(a0_k+tk*delta_a0_k,a1_k+tk*delta_a1_k,x,y)
        RHS = f(a0_k,a1_k,x,y)+alpha*tk*(-delta_a0_k*delta_a0_k-delta_a1_k*delta_a1_k)
        if LHS<RHS:
            break
        tk  = beta*tk
    
    # 3. akp1=ak+tk*Delta ak
    a0_kp1=a0_k+tk*delta_a0_k
    a1_kp1=a1_k+tk*delta_a1_k
    
    if np.linalg.norm(np.array((a0_kp1-a0_k,a1_kp1-a1_k)))<eps:
        print("GD converges at iter=",iter)
        break
    
    iter=iter+1
    a0_k,a1_k=a0_kp1,a1_kp1

print("a0=",a0_k)
print("a1=",a1_k)